package de.pdwexpress.parser.afgc_atgenexpress;

import de.ipk.pdwexpress.peobjects.*;

import java.io.*;

import org.apache.log4j.Logger;

/**
 * Parses a  _README.txt file, which represents a hybridisation. This file contains data about the whole
 * experiment (so, for more hybridisations several _README.txt files exist whith the same data about the experiment)
 * and about the sample etc. (specific for each hybridisation).
 * The experiment object is created externally by AFGC_AtGenExpress_ImportController. Here, the hybridisation object is
 * created as well as related data. This can be retrieved by getXXX()-methods.<br/>
 *
 * This readmeParser parses  a  _README.txt file for AFGC as well as AtGenExpress experiments.
 *
 * @author Andreas Stephanik
 */
public class AFGC_AtGenExpress_ReadmeParser {

    /**
     * The absolute directory where all files of an expression set (experiment) are located
     */
    String expressionSetDirectory;

    /**
     * The absolute path to the README file, which contains data about experiment (redundant for each hyb) and hyb
     */
    String absolutePath;

    BufferedReader br;

    protected Logger logger;

    /**
     * Is created externally by AFGC_AtGenExpress_ImportController
     */
    PE_EXPERIMENTS experiment;

    /**
     * Array design etc.
     */
    PE_ARRAYS array;

    /**
     * Hybridisation represented by this _README.txt file
     */
    PE_HYBRIDISATIONS hybridisation;

    /**
     * Has to be related to Hybridisation externally (foreign key)
     */
    PE_BIOMATERIALS biomaterial_1;

    /**
     * Only for AFGC
     */
    PE_BIOMATERIALS biomaterial_2;


    /**
     * Cy3 or Cy5 for AFGC or e.g. Biotin for AtGenExpress.
     * Has to be related to result set externally (after parsing and saving the according result set)
     */
    String sampleLabel_1;

    /**
     * Only for AFGC (Cy3 or Cy5)
     */
    String sampleLabel_2;


    /**
     * Flag if the data is from AFGC or AtGenExpress
     */
    boolean isAFGCFlag = false;

    /**
     * Absolute file name of the file with original data.
     * This is the .cel file with raw data from Affymetrix or
     * the original normalized data for AFGC microarrays.
     */
    String originalDataFileName;

    /**
     * Absolute file name with normalized data used a common TAIR normalization method.
     */
    String tairNormalizedDataFileName;


    // Further specific AFGC lines
    String afgcOriginalNormalizationMethod;
    String afgcOriginalNormalizationFactor;
    String afgcOriginalDataANOVASector;
    String afgcOriginalDataANOVAPlate;

    String tairNormalizationMethod; // also for AtGenExpress
    String afgcTAIRNormalizationFactor;
    String afgcTAIRDataANOVASector;
    String afgcTAIRDataANOVAPlate;

    // Further specific AtGenExpress lines
    String atgeScalingFactor;
    String atgeCentralMinusAvg;
    String atgeCentralMinusCount;
    String atgeCornerPlusAvg;
    String atgeCornerPlusCount;
    String atgeCornerMinusAvg;
    String atgeCornerMinusCount;

    // The original experiment name is extended by namespace 
    String expName;

    public static final String AFGC_TAIR = "AFGC:TAIR";

    public static final String ATGENEXPRESS_TAIR = "AtGenExpress:TAIR";


    public AFGC_AtGenExpress_ReadmeParser(String expressionSetDirectory, String absolutePath, PE_EXPERIMENTS experiment) throws FileNotFoundException {
        this.expressionSetDirectory = expressionSetDirectory;
        this.absolutePath = absolutePath;
        this.experiment = experiment;
        br = new BufferedReader(new FileReader(new File(absolutePath)));
        logger = Logger.getLogger(this.getClass());
        hybridisation = new PE_HYBRIDISATIONS(); // this _README.txt file represents a new hybridisation
        array = new PE_ARRAYS();
    }

    /**
     * Closes the internal BufferedReader.
     */
    public void closeParser() {
        try {
            br.close();
        }
        catch (Exception e) {
            logger.error("AFGC_AtGenExpress_ReadmeParser.closeParser():" + e);
        }
    }

    public PE_HYBRIDISATIONS parseAndImportHybData() throws Exception {
        parseTAIRAccession(); // store the TAIR accession number temporarily
        parseExperiment();
        parseRest();
        checkExperimentName();
        experiment.setOBJ_Name(expName);
        if(isAFGCFlag) {
            experiment.setDataSource(AFGC_TAIR);
        }
        else {
            experiment.setDataSource(ATGENEXPRESS_TAIR);
        }
        return hybridisation;
    }

    public PE_ARRAYS getArray() {
        return array;
    }

    public PE_HYBRIDISATIONS getHybridisation() {
        return hybridisation;
    }

    // getter methods
    public PE_BIOMATERIALS getBiomaterial_1() {
        return biomaterial_1;
    }

    public PE_BIOMATERIALS getBiomaterial_2() {
        return biomaterial_2;
    }

    /**
     * Cy3 or Cy5 for AFGC or e.g. Biotin for AtGenExpress.
     * Has to be related to result set externally (after parsing and saving the according result set)
     */
    public String getSampleLabel_1() {
        return sampleLabel_1;
    }

    public String getSampleLabel_2() {
        return sampleLabel_2;
    }

    public String getOriginalDataFileName() {
        return originalDataFileName;
    }

    public String getTairNormalizedDataFileName() {
        return tairNormalizedDataFileName;
    }

    public String getTairNormalizationMethod() {
        return tairNormalizationMethod;
    }

    public boolean isAFGCFlag() {
        return isAFGCFlag;
    }

    public String getAfgcOriginalNormalizationMethod() {
        return afgcOriginalNormalizationMethod;
    }

    public String getAfgcOriginalNormalizationFactor() {
        return afgcOriginalNormalizationFactor;
    }




    void parseTAIRAccession() throws Exception {
        String line = br.readLine();
        if(line != null) {
            if( line.startsWith("TAIR Accession =") ) { // first line of the file should start with that TAIR Accession
                String hybAcc = line.substring("TAIR Accession =".length()).trim();
                hybridisation.setOBJ_Identifier(hybAcc, PE_ExternDBEnum.TAIR);
                hybridisation.setOBJ_Name(hybAcc, hybAcc, PE_ExternDBEnum.TAIR);
                return;
            }
            else
                new Exception("Did not find expected data in " + absolutePath + ". Expected 'TAIR Accession ='");
        }
        throw new Exception("No data found in " + absolutePath + ". Expected 'TAIR Accession ='");
    }


    void parseExperiment() throws Exception {
        String line = br.readLine();
        while(line != null) {

            if( line.startsWith( "Experiment Submission Number =" ) ) {
                String expSubNumber = line.substring("Experiment Submission Number =".length()).trim() ;
                if(experiment.getOBJ_Identifier() == null)
                    experiment.setOBJ_Identifier("ExperimentSubmissionNumber:" + expSubNumber, PE_ExternDBEnum.TAIR); // Use the submission number as experiment identifier. This identifier is used in the ExperimentsSet_<ExpId> directory name
                //else if(!expSubNumber.equals(experiment.getOBJ_Identifier()))
                //    throw new Exception("Experiment identifier " + expSubNumber + " in hyb data file " + absolutePath + " is not equal to the previously set identifier " + experiment.getOBJ_Identifier());
            }

            else if( line.startsWith( "Experiment Name =" ) ) {
                expName = line.substring("Experiment Name =".length()).trim() ;
                if(experiment.getOBJ_Name() == null) {
                    //int index = expName.indexOf("ATGENEXPRESS:");
                    //if(index != -1) {
                    //    String str = expName.substring(index + "ATGENEXPRESS:".length()).trim();
                    //    experiment.setOBJ_Name(str, str, PE_ExternDBEnum.AtGenExpress);
                    //}
                    //else
                    //experiment.setOBJ_Name(expName); // is done later in parseAndImportHybData and checkExperimentName after parsing all, in order to set a prefix AFGC or ATGENEXPRESS
                }
                //else if( !expName.equals(experiment.getOBJ_Name()) )
                //    throw new Exception("Experiment name '" + expName + "' differs from previously set name '" + experiment.getOBJ_Name() + "'");

            }

            else if( line.startsWith("Experiment Description =") ) {
                // todo: could check previously set experiment description, as for Experiment Name
                if(experiment.getOBJ_Description() == null)
                    experiment.setOBJ_Description(line.substring("Experiment Description =".length()).trim().replaceAll("'", "''")); // single quote has to be replaced by double single quote
            }
            else if( line.startsWith("Experimenter =") ) {
                experiment.setExperimenter( line.substring("Experimenter =".length()).trim() );
                return;
            }
            line = br.readLine();
        }
        throw new Exception("Did not find expected data in " + absolutePath + ". Expected 'Experiment Submission Number =' and 'Experiment Name =' and 'Experiment Description =' and 'Experimenter ='");
    }


    void parseRest() throws IOException {
        String line = br.readLine();
        while(line != null) {

            if(line.startsWith("Slide Name =")) {
                // todo: is ignored at the moment, should be membran
                line.substring("Slide Name =".length()).trim() ;
            }

            else if(line.startsWith("Slide Description =")) {
                // todo: is ignored at the moment, should be membran
                line.substring("Slide Description =".length()).trim() ;
            }

            else if(line.startsWith("Name of the file containing raw data =")) { // specific for AtGenExpress
                originalDataFileName = expressionSetDirectory + line.substring("Name of the file containing raw data =".length()).trim() ;
            }

            else if(line.startsWith("Name of the file containing normalized data =")) { // specific for AtGenExpress
                tairNormalizedDataFileName = expressionSetDirectory + line.substring("Name of the file containing normalized data =".length()).trim() ;
            }
            else if(line.startsWith("TAIR name of the file containing normalized data =")) { // specific for AtGenExpress
                tairNormalizedDataFileName = expressionSetDirectory + line.substring("TAIR name of the file containing normalized data =".length()).trim() ;
            }

            else if(line.startsWith("AFGC exptid =")) {  // only for AFGC
                hybridisation.setOBJ_Identifier(AFGC_TAIR + ":Hyb:" + line.substring("AFGC exptid =".length()).trim(), PE_ExternDBEnum.AFGC);
                isAFGCFlag = true;
            }

            else if(line.startsWith("Array Design =")) {
                String design = line.substring("Array Design =".length()).trim();
                if(isAFGCFlag) {
                    array.setOBJ_Identifier(AFGC_TAIR + ":Array:" + design, PE_ExternDBEnum.AFGC);
                    array.setOBJ_Name(design, design, PE_ExternDBEnum.AFGC);
                }
                else {
                    array.setOBJ_Identifier(ATGENEXPRESS_TAIR + ":Array:" + design, PE_ExternDBEnum.Affymetrix);
                    array.setOBJ_Name(design, design, PE_ExternDBEnum.Affymetrix);
                }
            }


            else if(line.startsWith("Samples:")) { // only for AFGC (2 samples)
                parseAFGCSamplesDetails(br);
                isAFGCFlag = true; // sicher ist sicher
            }

            else if(line.startsWith("Sample:")) { // only for AtGenExpress (1 sample)
                parseAtGenExpressSampleDetail(line); // the sample details are in the same line
            }

            else if(line.startsWith("Scanner =")) { // obviously only for AtGenExpress
                line.substring("Scanner =".length()).trim();
            }

            else if(line.startsWith("Scanning Software & version =")) {
                line.substring("Scanning Software & version =".length()).trim();
            }

            else if(line.startsWith("Normalization method =")) {  // only for AtGenExpress, for AFGC we have the original normalized and the TAIR normalized
                tairNormalizationMethod = line.substring("Normalization method =".length()).trim();
            }

            else if(line.startsWith("File containing the original normalized data =")) { // AFGC
                originalDataFileName = expressionSetDirectory + line.substring("File containing the original normalized data =".length()).trim();
                isAFGCFlag = true;  // sicher ist sicher
            }

            //  some AFGC specific lines
            else if(line.startsWith("Original normalization method =")) // AFGC
                afgcOriginalNormalizationMethod = line.substring("Original normalization method =".length()).trim();

            else if(line.startsWith("Original normalization factor =")) // AFGC
                afgcOriginalNormalizationFactor = line.substring("Original normalization factor =".length()).trim();

            else if(line.startsWith("File containing TAIR-normalized data =")) // AFGC
                tairNormalizedDataFileName = expressionSetDirectory + line.substring("File containing TAIR-normalized data =".length()).trim();

            else if(line.startsWith("TAIR normalization method =")) // AFGC
                tairNormalizationMethod = line.substring("TAIR normalization method =".length()).trim();

            else if(line.startsWith("TAIR normalization factor =")) // AFGC
                afgcTAIRNormalizationFactor = line.substring("TAIR normalization factor =".length()).trim();

                // Further AtGenExpress specific data
            else if(line.startsWith("Normalization method ="))
                tairNormalizationMethod = line.substring("Normalization method =".length()).trim();

            else if(line.startsWith("Scaling factor ="))
                atgeScalingFactor = line.substring("Scaling factor =".length()).trim();

            else if(line.startsWith("Central- Avg ="))
                atgeCentralMinusAvg = line.substring("Central- Avg =".length()).trim();

            else if(line.startsWith("Central- Count ="))
                atgeCentralMinusCount = line.substring("Central- Count =".length()).trim();

            else if(line.startsWith("Corner+ Avg ="))
                atgeCornerPlusAvg = line.substring("Corner+ Avg =".length()).trim();

            else if(line.startsWith("Corner+ Count ="))
                atgeCornerPlusCount = line.substring("Corner+ Count =".length()).trim();

            else if(line.startsWith("Corner- Avg ="))
                atgeCornerMinusAvg = line.substring("Corner- Avg =".length()).trim();

            else if(line.startsWith("Corner- Count"))
                atgeCornerMinusCount = line.substring("Corner- Count".length()).trim();

            line = br.readLine();
        }
        new Exception("Did not find expected data in " + absolutePath + ". "+
                "Expected 'Slide Description =' or "+
                "'Array Design =' or" +
                "'Slide Name =' or" +
                "'Samples:' ...");
    }




    /**
     * Creates two samples for AFGC
     * @param br
     */
    void parseAFGCSamplesDetails(BufferedReader br) throws IOException {
        // two lines with details of Cy3 and Cy5 expected
        String line = br.readLine();
        if(line != null) {
            // first sample
            PE_BIOMATERIALS biomaterial = new PE_BIOMATERIALS();
            StringBuffer identifier = new StringBuffer(); // we can not use the name as identifier, cos in some experiments they use the same name for different materials, what can be seen in description; so, we add the hyb and channel number to the name
            StringBuffer label = new StringBuffer(); // Cy3 or Cy5
            parseAFGCSampleDetails(line, biomaterial, identifier, label);
            biomaterial_1 = biomaterial;
            biomaterial_1.setOBJ_Identifier("Hyb:" + hybridisation.getOBJ_Identifier() + ":CH1:" + identifier, PE_ExternDBEnum.AFGC);
            sampleLabel_1 = label.toString(); // Cy3 or Cy5

            // second sample
            line = br.readLine();
            if(line != null) {
                biomaterial = new PE_BIOMATERIALS();
                identifier = new StringBuffer();
                label = new StringBuffer(); // Cy3 or Cy5
                parseAFGCSampleDetails(line, biomaterial, identifier, label);
                biomaterial_2 = biomaterial;
                biomaterial_2.setOBJ_Identifier("Hyb:" + hybridisation.getOBJ_Identifier() + ":CH2:" + identifier, PE_ExternDBEnum.AFGC);
                sampleLabel_2 = label.toString();  // Cy3 or Cy5
            }
        }
        new Exception("Did not find expected sample data in " + absolutePath + ". Expected sample line starts with 'Name ='");
    }


    /**
     * Parses the data of one sample. The structure of a sample line has to be the same.
     * 	<code> Name = 7202_Cy3; Description = WT seedlings, 1 uM IAA for 30 min.; Tissue = whole plant; Development stage = seedling; Channel number = 1; Label = Cy3; test </code>
     *
     * @param line
     * @param label  The label Cy3 or Cy5, is returned as call by reference (String does not work as call by reference! if String then as array: String[0])
     */
    void parseAFGCSampleDetails(String line, PE_BIOMATERIALS biomaterial, StringBuffer identifier, StringBuffer label) {
        int index = line.indexOf("Name =");
        line = line.substring(index + "Name =".length()); // the rest of the line starting with the name
        index = line.indexOf(";");
        String name = line.substring(0, index).trim(); // the name till ;
        //biomaterial.setOBJ_Identifier(name, PE_ExternDBEnum.AFGC);
        identifier.append(name); // return using call by reference, cos we add the hyb and channel number to it
        biomaterial.setOBJ_Name( name );

        index = line.indexOf("Description =");
        line = line.substring(index + "Description =".length()); // the rest of the line starting with the descrption
        index = line.indexOf(";");
        biomaterial.setOBJ_Description( line.substring(0, index).trim() ); // the description could be parsed for more details

        index = line.indexOf("Tissue =");
        line = line.substring(index + "Tissue =".length() );
        index = line.indexOf(";");
        biomaterial.addOrganismPart( line.substring(0, index).trim() );

        index = line.indexOf("Development stage =");
        line = line.substring(index + "Development stage =".length() );
        index = line.indexOf(";");
        biomaterial.addDevelopmentalStage( line.substring(0, index).trim() );

        index = line.indexOf("Channel number =");
        line = line.substring(index + "Channel number =".length() );
        index = line.indexOf(";");
        // skipped

        index = line.indexOf("Label ="); // Cy3 or Cy5
        line = line.substring(index + "Label =".length() );
        index = line.indexOf(";");
        label.append( line.substring(0, index).trim() ); // is returned as call by reference (String does not work as call by reference! if String then as array: String[0])

        line = line.substring(index+1); // control or test after the ;
        if(line.trim().length() > 0)   // control type is not alwyas given
            biomaterial.addMatControlType(line);
    }


    void parseAtGenExpressSampleDetail(String line) {
        int index = line.indexOf("Name =");
        line = line.substring(index + "Name =".length()); // the rest of the line starting with the name
        index = line.indexOf(";"); // the name should end with a ;
        biomaterial_1 = new PE_BIOMATERIALS();
        biomaterial_1.addOrganism("Arabidopsis thaliana", "3702", PE_ExternDBEnum.NCBI_Taxonomy);
        String name = line.substring(0, index).trim(); // the name till ;
        biomaterial_1.setOBJ_Identifier(ATGENEXPRESS_TAIR + ":BioMaterial:" + name, PE_ExternDBEnum.AtGenExpress);
        biomaterial_1.setOBJ_Name(name, name, PE_ExternDBEnum.AtGenExpress);

        index = line.indexOf("Treatment description =");
        line = line.substring(index + "Treatment description =".length()); // the rest of the line starting behind Treatment description =
        index = line.indexOf(";");
        biomaterial_1.setOBJ_Description( line.substring(0, index).trim() ); // the description could be parsed for more details
        biomaterial_1.addClinicalTreatment( line.substring(0, index).trim() ); // todo: I am not sure why use it here, maybe needing the add-method instead the set-method

        index = line.indexOf("Anatomy description =");
        line = line.substring(index + "Anatomy description =".length());
        index = line.indexOf(";");
        biomaterial_1.addOrganismPart( line.substring(0, index).trim() );

        index = line.indexOf("Development stage =");
        line = line.substring(index + "Development stage =".length() );
        index = line.indexOf(";");
        biomaterial_1.addDevelopmentalStage( line.substring(0, index).trim() );

        index = line.indexOf("developmental stage "); // sometimes the arabidopsis developmental stage code is added
        if(index != -1) {
            line = line.substring(index + "developmental stage ".length() );
            index = line.indexOf(";");
            String boyes = line.substring(0, index).trim();
            biomaterial_1.addDevelopmentalStage(boyes, boyes, PE_ExternDBEnum.Boyes_Key_Growth_Stages );
        }

        index = line.indexOf("Label ="); // biotin
        line = line.substring(index + "Label =".length() );
        index = line.indexOf(";");
        sampleLabel_1 = line.substring(0, index).trim(); // biotin

        index = line.indexOf("Sample type =");
        line = line.substring(index + "Sample type =".length() );
        biomaterial_1.addMatControlType( line.substring(0).trim() );
    }

    void checkExperimentName() {
        if(isAFGCFlag) {
            if( !expName.startsWith("AFGC") && !expName.startsWith("afgc") ) {
                expName = "AFGC: " + expName;
            }
        }
        else {
            if( !expName.startsWith("ATGENEXPRESS") && !expName.startsWith("atgenexpress") ) {
                expName = "ATGENEXPRESS: " + expName;
            }
        }
    }
}
