package de.pdwexpress.parser.afgc_atgenexpress;

import de.ipk.pdwexpress.peobjects.PE_BIOSEQUENCES;
import de.ipk.pdwexpress.peobjects.PE_RESULTSETS;
import de.ipk.pdwexpress.peobjects.PE_ExternDBEnum;
import de.ipk.pdwexpress.peobjects.PE_EXPERIMENTS;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.PreparedStatement;
import java.util.HashMap;
import java.util.Set;
import java.util.Iterator;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;

/**
 * Parses the text csv files with expression data, creates result set
 * and saves them into pdw_express_stage.
 *
 * @author Andreas Stephanik
 */
public class AFGC_AtGenExpress_ResultSetParser {

    /**
     * The absolute path to the _original.txt file, which contains original expression data (AFGC or .cel files for AtGenExpress).
     */
    String absolutePath;

    /**
     * Reads the files line by line
     */
    BufferedReader br;

    Connection connection;

    Logger logger;

    /**
     * Parser that read experiment ans sample annotations before.
     * So, material etc. was created by that parser.
     */
    AFGC_AtGenExpress_ReadmeParser readmeParser;

    /**
     * Key is the name of the column (result sets name) and
     * the value is the primary key of the DB of the result set.
     */
    HashMap<String,String> resultSetMaterial1PksMap;

    /**
     * Same as for resultSetMaterial1PksMap, but this here
     * is for the second channel for AFGC.
     */
    HashMap<String,String> resultSetMaterial2PksMap;

    /**
     * Key is the name of the column (result sets name) and
     * the value is the index (Integer, starting at 0) of the column.
     * The index is used in insertResultSetsValues(), where a splitted parsed line
     * is available as a String-array. So, the data of each column can be
     * accessed directly by the index of the column.
     */
    HashMap<String,Integer> resultSetColumnIndexesMap;

    /**
     * Following ints are the indexes of the columns regarding the string array, which comes from parseLine().
     * So, the first column (ARRAY_ELEMENT) has the index 0.
     * Using these indexes the corresponding values can be accessed.
     * The method searchForColumnIndexes() can be used for searching the indexes by column header strings.
     * At the moment we assume that the column order is fixed.
     * @deprecated  still partially used for AFGC
     */
    int array_element = 0, locus = 1, suid = 2, ch1i_mean = 4, ch1b_mean = 9, ch1b_median = 10, ch2i_mean = 13, ch2b_mean = 19, ch2b_median = 20;

    /**
     * The index of the column where the Affy biosequence is located
     */
    int affyBiosequenceColumn = -1;

    /**
     * Name of the result set for Affymetrix.
     * Currently, the value is not changed, e.g. for another column header.
     */
    String affyRSetName = "Signal";



    String statement = "insert into RESULT_DATA_T  (RSD_OBJ_ID_RST,  RSD_OBJ_ID,  RSD_VALUE) " +
            "values (? , ? ,  ?)";

    /**
     * Is opened in parseFileAndSaveData but also used in insertResultSetsValues, which is called
     * within a loop.
     */
    private PreparedStatement pstmt;

    /**
     * We need some information in while parsing the normalised data, e.g. the FKs of the result sets of the original data
     */
    AFGC_AtGenExpress_ResultSetParser originalDataResultSetParser;

    /**
     * Flag, if we parse original or normalised data.
     */
    boolean normalisedData = false;


    /**
     * Constructor for importing original data.
     *
     * @param absolutePath
     * @param readmeParser
     * @param connection
     * @throws FileNotFoundException
     */
    public AFGC_AtGenExpress_ResultSetParser(String absolutePath, AFGC_AtGenExpress_ReadmeParser readmeParser, Connection connection) throws FileNotFoundException {
        this(absolutePath, readmeParser, null, connection);
    }

    /**
     * Constructor for importing normalised data (AFGC).
     *
     * @param absolutePath
     * @param readmeParser
     * @param originalDataResultSetParser   The parser used to import original data. The result sets of original data are needed for setting FKs of the normalised result set to them.
     * @param connection
     * @throws FileNotFoundException
     */
    public AFGC_AtGenExpress_ResultSetParser(String absolutePath, AFGC_AtGenExpress_ReadmeParser readmeParser, AFGC_AtGenExpress_ResultSetParser originalDataResultSetParser, Connection connection) throws FileNotFoundException {
        System.out.println("Data file:\t" + absolutePath);
        this.absolutePath = absolutePath;
        br = new BufferedReader(new FileReader(absolutePath));
        this.readmeParser = readmeParser;
        if(originalDataResultSetParser != null) {
            this.originalDataResultSetParser = originalDataResultSetParser; // original data was already parsed, so, now parsing the normalised data
            normalisedData = true;
        }
        this.connection = connection;
        logger = Logger.getLogger(this.getClass());
    }



    /**
     * First, take first line with header (column names) and populate resultSetColumnIndexesMap.
     * Create result sets for each definined column (in createAFGCResultSets).
     * Then, parse the data lines with biosequence and result set information and
     * save them according to the result set definitions.
     *
     * @throws Exception any Exception (IOException, SQLException or self thrown Exception)
     */
    public void parseFileAndSaveData() throws Exception {
        String line = br.readLine();
        if(line != null) {
            pstmt = connection.prepareStatement(statement); // in searchForColumnIndexes() ( guessAtGenExpressFormat() respectively) the prepareStatement could be used if the first line has already expression data (no column header)
            searchForColumnIndexes(parseLine(line)); // first line with column header (not always in AtGenExpress!), populate resultSetColumnIndexesMap
            if(readmeParser.isAFGCFlag())
                createAFGCResultSets();
            else if(affyBiosequenceColumn != -1) {// if a column for the biosequence was found
                if (resultSetMaterial1PksMap == null)  // if the result set for affymetrix was not already created by invoking createAtGenExpressesultSet from guessAtGenExpressFormat; this happens, if the first line has already data
                    createAtGenExpressesultSet();
            }
            else throw new Exception("AFGC_AtGenExpress_ResultSetParser: No result data found in " + absolutePath); // no suitable data for AtGenExpress found. throw an Exception???
            line = br.readLine(); // next line with data
        }
        int i = 0; // zum Testen von nur ein paar lines mit i < 10
        // now, iterate thru the data lines and insert the Biosequences and the Signal values
        while(line != null) { // zum Testen von nur ein paar lines mit:  && i < 10
            //System.out.println(createdCSVLine(parseLine(line)));
            if(line.trim().length() > 0) {
                String[] values = parseLine(line);
                //System.out.println("");
                //logger.info("Parsed line:" + createdCSVLine(values));  // just for testing
                insertResultSetsValues(values);
            }
            line = br.readLine();
            i++;
        }
        if(pstmt != null)
            pstmt.close();
    }

    public HashMap<String, String> getResultSetMaterial1PksMap() {
        return resultSetMaterial1PksMap;
    }

    public HashMap<String, String> getResultSetMaterial2PksMap() {
        return resultSetMaterial2PksMap;
    }

    /**
     * Parses a line and splits the line by tab-separator.
     *
     * @param line  string line.
     * @return  String-array with data of each column.
     */
    String[] parseLine(String line) {
        /*
        ArrayList<String> values = new ArrayList<String>();
        StringTokenizer tokenizer = new StringTokenizer(line , "\t");
        while(tokenizer.hasMoreTokens()) {
            values.add(tokenizer.nextToken());
        }
        String[] ret = new String[values.size()];
        for (int i = 0; i < values.size(); i++) {
            ret[i] = values.get(i);
        }
        return ret;
        */
        return  line.split("\t", -1);
    }


    /**
     * Looks up for the right index of each column.
     *
     * @param columnHeader  The first line of the TSV file  ARRAY_ELEMENT	LOCUS	SPOT	SUID	CH1I_MEAN ...
     */
    void searchForColumnIndexes(String[] columnHeader) throws SQLException, IOException {
        resultSetColumnIndexesMap = new HashMap<String, Integer>(); // generic approach, where all column names and their indexes are stored
        if(readmeParser.isAFGCFlag()) {
            for(int i = 0; i < columnHeader.length; i++) {
                /*
                if("CH1I_MEAN".equals(columnHeader[i])) { // found the
                    ch1i_mean = i;
                }
                else if("CH1B_MEAN".equals(columnHeader[i])) {
                    ch1i_mean = i;
                }
                // and so on
                */
                resultSetColumnIndexesMap.put(columnHeader[i], i); // column name and its index
            }
        }
        else {
            // different formats for AtGenExpress and .cel for original
            guessAtGenExpressFormat(columnHeader);
        }
    }


    /**
     * Definition of the result sets.
     * Each result set has to have following elements:
     * Column name (result set name), result set type (e.g.RawDataDensity or RawDataBackground), normalization method (from readme parser) or null, description
     */
    void createAFGCResultSets() throws SQLException {

        if( !normalisedData ) { // original data file

            // All result sets (foreground, background) for sample 1
            resultSetMaterial1PksMap = new HashMap<String, String>();

            // Sample 1, foreground
            String resultSetPK = createResultSet("CH1I_MEAN",  // The identifier; here "CH1I_MEAN" has to fit to the key of the HashMap resultSetMaterial1PksMap !!!
                    "CH1I_MEAN",
                    "RawData Density",
                    null,
                    "Mean spot pixel intensity at Channel 1 (usually 532 nm).",
                    readmeParser.getSampleLabel_1(),
                    readmeParser.getBiomaterial_1().getObj_id(),
                    null);

            resultSetMaterial1PksMap.put("CH1I_MEAN", resultSetPK); // the first element is the name, e.g. "CH1I_MEAN"; this name is used as key in the HashMap result

            // Sample 1, background
            resultSetPK = createResultSet("CH1B_MEAN",
                    "CH1B_MEAN",
                    "RawData Background",
                    null,
                    "Mean spot background in Channel 1 (usually 532 nm).",
                    readmeParser.getSampleLabel_1(),
                    readmeParser.getBiomaterial_1().getObj_id(),
                    null);

            resultSetMaterial1PksMap.put("CH1B_MEAN", resultSetPK);



            // All result sets (foreground, background) for sample 2
            resultSetMaterial2PksMap = new HashMap<String, String>();


            // Sample 2, foreground
            resultSetPK = createResultSet("CH2I_MEAN",
                    "CH2I_MEAN",
                    "RawData Density",
                    null,
                    "Mean spot pixel intensity at Channel 2 (usually 532 nm).",
                    readmeParser.getSampleLabel_2(),
                    readmeParser.getBiomaterial_2().getObj_id(),
                    null);

            resultSetMaterial2PksMap.put("CH2I_MEAN", resultSetPK); // the first element is the name, e.g. "CH2I_MEAN"; this name is used as key in the HashMap result

            // Sample 2, background
            resultSetPK = createResultSet("CH2B_MEAN",
                    "CH2B_MEAN",
                    "RawData Background",
                    null,
                    "Mean spot background in Channel 2 (usually 532 nm).",
                    readmeParser.getSampleLabel_2(),
                    readmeParser.getBiomaterial_2().getObj_id(),
                    null);

            resultSetMaterial2PksMap.put("CH2B_MEAN", resultSetPK);
        }
        else { // normalised data

            resultSetMaterial1PksMap = new HashMap<String, String>();

            String resultSetPK = createResultSet("FOLD_CHANGE", // creates the result set for TAIR-analysed data
                    "FOLD_CHANGE",
                    "Fold change",
                    null,
                    "How many times the expression signal for a given transcript increased or decreased compared to the control.",
                    null,
                    originalDataResultSetParser.getResultSetMaterial1PksMap().get("CH1I_MEAN"),  // result set PK of original data channel 1
                    originalDataResultSetParser.getResultSetMaterial2PksMap().get("CH2I_MEAN")); // result set PK of original data channel 2

            resultSetMaterial1PksMap.put("FOLD_CHANGE", resultSetPK); // FOLD_CHANGE has to be in the header of the tair.txt file!!! The same string FOLD_CHANGE is hopefully in resultSetColumnIndexesMap, which is populatetd in searchForColumnIndexes

            resultSetMaterial2PksMap = null;
        }

    }


    /**
     * Definition of the result sets.
     * Each result set has to have following elements:
     * Column name (result set name), result set type (e.g.RawDataDensity or RawDataBackground), normalization method (from readme parser) or null, description
     */
    void createAtGenExpressesultSet() throws SQLException {
        // resultSetMaterial1Pks could already be created, when this method was invoked by guessAtGenExpressFormat(); BUT then, this method should not be invoked again, cos resultSetMaterial1Pks is not null
        resultSetMaterial1PksMap = new HashMap<String, String>();

        String resultSetPK = null;
        resultSetPK = createResultSet(affyRSetName,
                affyRSetName,
                "NormalisedData",
                readmeParser.getTairNormalizationMethod(),
                "",
                readmeParser.getSampleLabel_1(),
                readmeParser.getBiomaterial_1().getObj_id(), null);

        resultSetMaterial1PksMap.put(affyRSetName, resultSetPK); // the first element is the identifier, which is mapped to the PK of the result set
    }


    /**
     * Creates a PE_RESULTSETS object for each result set in resultSetsDictionary, saves
     * them into the DB and put the primary key into a HashMap, which is returned. The key
     * of the HashMap is the name of the result set, e.g. CH1I_MEAN.
     *
     * @param label             Used label, e.g. biotin for Affymetrix.
     * @param biomaterialOrFirstRSetPK     Either the PK of the corrsponding biomaterial, or the PK of the first result set for creating a derived result set.
     * @param secondRSetPK      The PK of the second result set for creating a derived result set, e.g. for fold change of 2-colors-arrays.
     *
     * @return A HasMap with result set names as key (e.g. CH1I_MEAN) and the DB primary key of the corresponding saved PE_RESULTSETS object
     */
    String createResultSet(String identifier, String name, String type, String normalisationMethod, String description, String label, String biomaterialOrFirstRSetPK, String secondRSetPK) throws SQLException {
        // elements of resultSetsDictionary: Column name (result set name), result set type, normalization method (from readme parser), description

        PE_RESULTSETS resultSet;

        resultSet = new PE_RESULTSETS();
        if(readmeParser.isAFGCFlag())
            resultSet.setOBJ_Identifier("AFGC:Hyb:"     + readmeParser.getHybridisation().getOBJ_Identifier() + ":ResultSet:" + identifier, PE_ExternDBEnum.AFGC);
        else
            resultSet.setOBJ_Identifier("AtGenExpress:" + readmeParser.getHybridisation().getOBJ_Identifier() + ":ResultSet:" + identifier, PE_ExternDBEnum.AtGenExpress);
        resultSet.setOBJ_Name(name); // result set name, taken from column name
        resultSet.setResultSetType(type); // result set type, e.g. RawDataDensity  or RawDataBackground
        if(normalisationMethod != null && normalisationMethod.length() > 0)  // normalization method, if available
            resultSet.setNormalizationDescriptionType(normalisationMethod);
        if(description != null && description.length() > 0)
            resultSet.setOBJ_Description(description);  // result set description
        if(label != null)
            resultSet.setLabelCompound(label); // sample label (Cy3 or cy5 for AFGC or Biotin for AtGenExpress)
        if(secondRSetPK == null)
            resultSet.setFK_RST_MAT_ID(biomaterialOrFirstRSetPK); // biomaterial PK
        else {
            resultSet.addFK_RST_RST_ID_PARENT(biomaterialOrFirstRSetPK); // result set PK of original data channel 1
            resultSet.addFK_RST_RST_ID_PARENT(secondRSetPK);  // result set PK of original data channel 2
        }
        resultSet.setFK_RST_HYB_ID(readmeParser.getHybridisation().getObj_id());
        return resultSet.saveObject();
    }






    /**
     * Takes the values of one line.
     * @param values
     * @throws SQLException
     */
    void insertResultSetsValues(String values[]) throws SQLException {
        // schleife zeilenweise ist bereits gegeben und wir haben alle values einer Zeile
        // dann Biosequenz holen (SUID und spter locus(TAIR) bei AFGC)
        // dann den entsprechenden value aus der Position im array values[] nehmen und im jeweiligen resultset (ch1i_mean etc.) mit dem Prepared-Statement ablegen, dazu muss der PK des resultsets vorher ermittelt werden
        PE_BIOSEQUENCES biosequence = new PE_BIOSEQUENCES(); // todo: maybe, the biosequence object can created ealier, e.g. after parsing the line. So, additional infos could be added, e.g. description of the biosequence, found in some columns in the files
        if(readmeParser.isAFGCFlag()) {
            String seqId = values[resultSetColumnIndexesMap.get("ARRAY_ELEMENT")].trim();
            if( "EMPTY".equalsIgnoreCase(seqId) || seqId.length() == 0)
                seqId = null; // we cannot use EMPTY as the sequence identifier

            String seqIdAgi = values[resultSetColumnIndexesMap.get("LOCUS")];
            if(seqIdAgi != null && seqIdAgi.trim().length() > 0) {
                biosequence.addOBJ_ExternID(seqIdAgi, seqIdAgi, PE_ExternDBEnum.AGI); // e.g. AT5G60390
            }
            else
                seqIdAgi = null; // an AGI identifier was not found

            //biosequence.setOBJ_Identifier(values[suid], PE_ExternDBEnum.SMD);  // deprecated, use the next line
            String seqIdSmd = values[resultSetColumnIndexesMap.get("SUID")];
            if(seqIdSmd != null && seqIdSmd.trim().length() > 0)
                biosequence.addOBJ_ExternID(seqIdSmd, seqIdSmd, PE_ExternDBEnum.SMD);
            else
                seqIdSmd = null;

            if(seqId != null) // seqId is null if EMPTY was found
                biosequence.setOBJ_Identifier(seqId, PE_ExternDBEnum.AFGC);
            else if(seqIdSmd != null)
                biosequence.setOBJ_Identifier(seqIdSmd, PE_ExternDBEnum.SMD);
            else if(seqIdAgi != null)
                biosequence.setOBJ_Identifier(seqIdAgi, PE_ExternDBEnum.AGI);
            else
                biosequence.setOBJ_Identifier("EMPTY", PE_ExternDBEnum.AFGC);

            //System.out.println("BioSequence-Identifier:" + biosequence.getOBJ_Identifier() + "\t\tseqId:" + seqId);

            if(seqIdAgi != null)
                biosequence.setOBJ_Name(seqIdAgi, seqIdAgi, PE_ExternDBEnum.AGI);
            else if(seqIdSmd != null)
                biosequence.setOBJ_Name(seqIdSmd, seqIdSmd, PE_ExternDBEnum.SMD);
        }
        else { // AtGenExpress
            biosequence.setOBJ_Identifier(values[affyBiosequenceColumn], PE_ExternDBEnum.Affymetrix); // e.g. AFFX-BioB-5_at
        }
        String bioSeqId = biosequence.saveObject();

        insertResultSetsValues(values, resultSetMaterial1PksMap, bioSeqId); // the HashMap resultSetMaterial1Pks should only contain 1 element for AtGenExpress, It can also be defined a result set for background (for AFGC) for each channel (material)
        if(readmeParser.isAFGCFlag() && resultSetMaterial2PksMap != null)
            insertResultSetsValues(values, resultSetMaterial2PksMap, bioSeqId); // the second channel for AFGC
    }


    /**
     * Inserts all desired values of a line according to the result sets. A line can contain values for more than one result set, e.g. foreground, background, sample 1, sample 2
     *
     * @param values
     * @param resultSetPKs
     * @param bioSeqId
     * @throws SQLException
     */
    void insertResultSetsValues(String values[], HashMap<String, String> resultSetPKs, String bioSeqId) throws SQLException {
        Set<String> resultSetNames = resultSetPKs.keySet();
        String rsetName;
        for (Iterator<String> iterator = resultSetNames.iterator(); iterator.hasNext();) {
            rsetName = iterator.next();  // the result set with name was defined in createAFGCResultSets
            //logger.info("ResultSet:" + rsetName + "\tBioSeq:" + bioSeqId + "\tValue:" + values[ columnIndexesMap.get(rsetName) ]);
            pstmt.setInt( 1, Integer.parseInt(resultSetPKs.get(rsetName)) );  // Pk of the already saved result set
            pstmt.setInt( 2, Integer.parseInt(bioSeqId) );    // PK of the already saved biosequence
            pstmt.setFloat( 3, Float.parseFloat(values[ resultSetColumnIndexesMap.get(rsetName) ]) );  // resultSetColumnIndexesMap manages the index of the column of the current result set; the key rsetName was defined in createAFGCResultSets or createAtGenExpressesultSet; resultSetColumnIndexesMap should contain only one result set (Signal) for AtGenExpress
            pstmt.execute();
        }
    }



    /**
     * Looks for the Biosequence column and the Signal column.
     * These two columns can be found in different formats. So, various approaches
     * are needed.
     * In some cases we do not find a Signal column, so we skip those files.
     *
     * @param columnHeader  First line of the TSV file, which has either a header with column names,
     *  or already the first data entry
     *
     * @throws SQLException
     * @throws IOException
     */
    void guessAtGenExpressFormat(String[] columnHeader) throws SQLException, IOException {
        logger.info("COLUMN Header length:" + columnHeader.length + " LINE:" + createdCSVLine(columnHeader));

        if( (columnHeader.length == 6 || columnHeader.length == 7) && !columnHeader[0].contains("=") ) { // skip line with 6 cells but content like: "BF="
            affyBiosequenceColumn = 0;  // the biosequence is in first column
            if( columnHeader[0].startsWith("AFFX-") ) {   // the first line is already a data line
                // ExpressionSet_ME00342:
                //
                // AFFX-BioB-5_at	20	20	50.7	Present	0.00256
                resultSetColumnIndexesMap.put(affyRSetName, 3);  // affyRSetName is Signal
                createAtGenExpressesultSet();    // the first line is already a data line, so, create the result set
                insertResultSetsValues(columnHeader);  // and insert the data of the first line
            }
            else {
                // ExpressionSet_ME00341:
                //
                // 	\t              "SF At+Bc MA 1-18-i-1 #3_Stat Pairs"	"SF At+Bc MA 1-18-i-1 #3_Stat Pairs Used"	"SF At+Bc MA 1-18-i-1 #3_Signal"	"SF At+Bc MA 1-18-i-1 #3_Detection" "SF At+Bc MA 1-18-i-1 #3_Detection p-value"	"Descriptions"
                // AFFX-BioB-5_at	20	                                    20	                                        235.3	                            P	                                0.003595	                                J04423 E coli bioB gene biotin synthetase  (-5, -M, -3 represent transcript regions 5 prime, Middle, and 3 prime respectively)
                affyRSetName = columnHeader[3];  // take the column header string where the signal is stored
                resultSetColumnIndexesMap.put(affyRSetName, 3);
            }
        }
        else {  // a real header with column names is given
            // Example 1:
            // Probe Set Name	Stat Pairs	Stat Pairs Used	    Signal	Detection	Detection p-value	Stat Common Pairs	Signal Log Ratio	Signal Log Ratio Low	Signal Log Ratio High	Change	Change p-value	Positive	Negative	Pairs	Pairs Used	Pairs InAvg	Pos Fraction	Log Avg	Pos/Neg	Avg Diff	Abs Call	Inc	Dec	Inc Ratio	Dec Ratio	Pos Change	Neg Change	Inc/Dec	DPos-DNeg Ratio	Log Avg Ratio Change	Diff Call	Avg Diff Change	B=A	Fold Change	Sort Score
            // AFFX-BioB-5_at	20	        20	                89	    P	        0.000972
            //
            //
            // Example2:
            // \t	Analysis Name	Probe Set Name	Stat Pairs	Stat Pairs Used	    Signal	Detection	Detection p-value	Stat Common Pairs	Signal Log Ratio	Signal Log Ratio Low	Signal Log Ratio High	Change	Change p-value	Positive	Negative	Pairs	Pairs Used	Pairs InAvg	Pos Fraction	Log Avg	Pos/Neg	Avg Diff	Abs Call	Inc	Dec	Inc Ratio	Dec Ratio	Pos Change	Neg Change	Inc/Dec	DPos-DNeg Ratio	Log Avg Ratio Change	Diff Call	Avg Diff Change	B=A	Fold Change	Sort Score
            //  1	co-0-1          AFFX-BioB-5_at	20          20	                115.3	P	        0.000754
            //
            //
            // Example3:
            // \t	            GROWTH CHAMBER #5_Stat Pairs	GROWTH CHAMBER #5_Stat Pairs Used	GROWTH CHAMBER #5_Signal	GROWTH CHAMBER #5_Detection	GROWTH CHAMBER #5_Detection p-value	GROWTH CHAMBER #5_Stat Common Pairs	GROWTH CHAMBER #5_Signal Log Ratio	GROWTH CHAMBER #5_Signal Log Ratio Low	GROWTH CHAMBER #5_Signal Log Ratio High	GROWTH CHAMBER #5_Change	GROWTH CHAMBER #5_Change p-value	Descriptions
            // <empty line>
            //  AFFX-BioB-5_at  20
            //
            String line = "";
            boolean signal = false; // until the column with signals (expression values) has been found
            for( int li = 0; li < 80; li++) { // check the first 80 lines, if necessary
                for(int i = 0; i < columnHeader.length; i++) {
                    if( !signal && (columnHeader[i].contains("Signal") || "Log Avg".equals(columnHeader[i])) ) { // look for expression levels; "Log Avg" is for ME00313
                        signal = true;
                        resultSetColumnIndexesMap.put(columnHeader[i], i);
                        affyRSetName = columnHeader[i];  // so, affyRSetName is "Signal"
                        if(affyBiosequenceColumn != -1)
                            return; // we have all what we need
                    }
                    else if( affyBiosequenceColumn == -1 && columnHeader[i].contains("Probe Set") ) {
                        affyBiosequenceColumn = i; // the biosequence is defined by column "Probe Set Name"
                    }
                    else if(columnHeader[i].startsWith("AFFX-")) { // we found Signal column but not the "Probe Set" column with (AFFX-), so, look for AFFX-  in following lines
                        affyBiosequenceColumn = i; // the column of the signal should already be set
                        createAtGenExpressesultSet();    // the first line is already a data line, so, create the result set
                        insertResultSetsValues(columnHeader);
                        if(signal)
                            return;  // return only if a column with expression levels was found
                    }
                }
                line = br.readLine(); // not all what we need found, so, read more lines
                while(line.trim().length() == 0) { // skip empty lines
                    line = br.readLine();
                    if(line == null)
                        return;
                }
                columnHeader = parseLine(line);
            }
        }
    }


    String createdCSVLine(String[] values) {
        StringBuffer sb = new StringBuffer();
        //logger.info("Lnge:" + values.length+ "\t");
        for(int i = 0; i < values.length; i++) {
            sb.append(values[i]);
            if(i < values.length-1)
                sb.append("\t");
        }
        return sb.toString();
    }



    void testParser() throws SQLException, IOException {
        String line = br.readLine();
        searchForColumnIndexes( parseLine(line) );
    }



    public static void main(String[] args) {
        try {
            BasicConfigurator.configure(); // log4j

            AFGC_AtGenExpress_ReadmeParser readmeParser = new AFGC_AtGenExpress_ReadmeParser("D:\\Downloads\\AtGenExpress\\Datasets\\ExpressionSet_ME00317\\",
                    "D:\\Downloads\\AtGenExpress\\Datasets\\ExpressionSet_ME00313\\10_README.txt",
                    new PE_EXPERIMENTS());

            AFGC_AtGenExpress_ResultSetParser resultSetParser = new AFGC_AtGenExpress_ResultSetParser("D:\\Downloads\\AtGenExpress\\Datasets\\ExpressionSet_ME00313\\10.txt",
                    readmeParser,
                    null);

            resultSetParser.testParser();
            //resultSetParser.parseFileAndSaveData();
            //resultSetParser.createAFGCResultSets();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }
}
