package de.pdwexpress.parser.common;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;


/**
 * Parser for data in MatrixFormat <br>
 * <br>
 * Structur:<br>
 * First Column: 'labels'<br>
 * Others Data (every Column is one Dataset)<br>
 * <br>
 * Parser assums that T has setter methodes for the 'labels'
 * <br>
 * Example: <br>
 * 		MatrixParser<PE_BIOMATERIALS> ap = new MatrixParser<PE_BIOMATERIALS>(new PE_BIOMATERIALS());<br>
 *		List<PE_BIOMATERIALS> allObj =  ap.parseFile(new File("C:\\matTest.TXT"));<br>
 *		<br>	
 *		if (allObj ==null){<br>
 *		System.err.println("Fatal error.");<br>
 *			System.exit(-1);<br>
 *		}<br>
 *		<br>				
 *		for (PE_BIOMATERIALS bm:allObj )<br>
 *			System.out.println(bm.getOBJ_Name()+", "+bm.getOrganismPart()+", "+bm.getTemperature()+", "+bm.getBedding());<br>
 *
 * @author Thomas
 *
 * @param <T>
 */
public class MatrixParser<T extends Object> {
	
	protected Logger logger;
	protected final static String SEPARATOR="\t";
	protected Class clazz;
	
	public MatrixParser(T t){
		logger = Logger.getLogger(this.getClass());
		this.clazz=t.getClass();		
	}

	
	/**
	 * Parses the file and returns a List of Objects (Type T).<br>
	 * First column contains the AttributeNames.<br>
	 * The parser asmues that T has a setMathod for the Attribut.<br>
	 * The following columns represents the objects.
	 * 
	 * @param sourceFile The file.
	 * 
	 * @return List of Objects. Never NULL.
	 * 
	 * @throws NoSuchMethodException
	 * 
	 * @throws IOException
	 */
	public List<T> parseFile(File sourceFile) throws NoSuchMethodException, IOException{
		
		
		logger.debug("open file "+sourceFile);
		FileReader fr = new FileReader(sourceFile);
		BufferedReader br = new BufferedReader(fr);
		
		// read first line
		String currentLine=null;
//		StringTokenizer tokenizer =null;
		

		// Read first line, which is not a comment
		try {
			do{
				currentLine=br.readLine();
				if (currentLine.startsWith("#"))
					logger.info("Commentline:"+currentLine);
			}while (currentLine.startsWith("#"));
		} catch (IOException e) {
			logger.error(" Error while reading from file "+sourceFile);
			e.printStackTrace();
			IOUtils.closeQuietly(fr);
			throw e;
		}
		
		// get ColumnCount
		int objectCount = splitLine(currentLine).size()-1;
		logger.info(objectCount+ " objects found ");
		
		// get the default Constructor of T
		final Constructor constructor = clazz.getConstructors()[0];
		
		// The Parameter to invoke the Constructor 
		// Object[0] => Empty because there are no Parameter
		final Object[] constructorParam =new Object[0]; 
		
		// create return List 
		logger.debug("Creating Objects of Type T");
		final List<T> allObj = new Vector<T>();
		
		// create T Objects within List
		try {
			
			// Objecte erstellen
			for (int i=0; i<objectCount;i++)
				allObj.add((T) constructor.newInstance(constructorParam ));
			
		} catch (Exception e) {
			String m="Error while creating Objects of Type T"+e;
			logger.error(m);
			e.printStackTrace();
			IOUtils.closeQuietly(fr);
			throw new NoSuchMethodException(m);
		}
		
		
		// read line by line
		try {
			do{	
				
				if (currentLine == null){
					logger.warn("file is empty");
					// file is empty
					break;
				}
				
				if (currentLine.startsWith("#")){
					logger.info("Commentline:"+currentLine);
					continue;
				}
				
				// parse line
				Iterator<String> lineIterator = splitLine(currentLine).iterator();		
				
				int tokenIndex=-1; // 
				String currentAttrName=null;
				Method method=null;
				
				// read the Elements
				while(lineIterator.hasNext()){
					tokenIndex++;
					
					if (tokenIndex ==0){
						method=null;
						currentAttrName=lineIterator.next();
						String setterMethodName="set"+currentAttrName;
						
						// get the Setter-Method
						logger.debug("searching for Method "+setterMethodName);
						try {
							method =clazz.getMethod(setterMethodName, String.class);
						} catch (Exception e) {
							String m="Error while searching for Method "+setterMethodName;
							logger.fatal(m);
							e.printStackTrace();
							IOUtils.closeQuietly(fr);
							throw new NoSuchMethodException(m);
						}						
						continue;						
					}
					
					// read value from file
					String value = (String) lineIterator.next();					
					logger.debug("value: "+value);
					if (value == null){
						logger.debug(currentAttrName+" is null");
						// It's not necessary to call setter for NULL
						continue;					
					}

						
					try {
						// call setter Metod
						method.invoke(allObj.get(tokenIndex-1), value);
					} catch (Exception e) {
						String m="Error while invoking Methode:"+method+ " for Attr:"+currentAttrName+ " with value:"+value;
						logger.fatal(m);
						e.printStackTrace();
						IOUtils.closeQuietly(fr);
						throw new  NoSuchMethodException(m);
					} 	

					
				}
				
				// read new line
			} while((currentLine=br.readLine()) != null);
			
						
		} catch (IOException e) {
			e.printStackTrace();
		}finally{
			IOUtils.closeQuietly(fr);
		}
		
				
		
		return allObj;
	}

	

	/**
	 * calls: splitLine(line,"\t")
	 * @param line
	 * @return
	 */
	public static List<String> splitLine(String line){
		return splitLine(line,"\t");
	}
	
	/**
	 * Parses the line. <br>
	 * Items within the delimiters are trimed.<br>
	 * Items with lengt 0 are replaced with NULL.<br>
	 * 
	 * @param line A String with delimiters.
	 * @param delimiter The delimiter.
	 * @return List of String or NULL values. Never NULL.
	 */
	public static List<String> splitLine(String line,String delimiter) {

		int delimPosStart=0;
		int delimPosEnd=-1;
		Vector<String> list = new Vector<String>();
		
		if (line.length()>0)
		do{
			delimPosStart=++delimPosEnd;
			
			delimPosEnd=line.indexOf(delimiter, delimPosStart);
			
			if (delimPosEnd==-1){
				delimPosEnd=line.length();
			}
			
			String item = line.substring(delimPosStart, delimPosEnd).trim();
			list.add((item.length()==0)?null:item);
			
						
		}while(delimPosEnd != line.length());
		
		return list;
	}	
}
